@@ -2,9 +2,13 @@
 
 """
 
+import os
+
 import datetime
 import logging
 import multiprocessing as mp
+import shutil
+import tempfile
 import uuid
 
 from pkg_resources import iter_entry_points
@@ -13,7 +17,7 @@ from collections import defaultdict
 import dtoolcore.utils
 
 
-__version__ = "3.14.0"
+__version__ = "3.17.0"
 
 logger = logging.getLogger(__name__)
 
@@ -35,32 +39,47 @@ def _get_identifier_and_item_properties(d):
 
 def _generate_storage_broker_lookup():
     """Return dictionary of available storage brokers."""
+    logger.debug("In _generate_storage_broker_lookup...")
     storage_broker_lookup = dict()
     for entrypoint in iter_entry_points("dtool.storage_brokers"):
         StorageBroker = entrypoint.load()
-        storage_broker_lookup[StorageBroker.key] = StorageBroker
+        key = StorageBroker.key
+        logger.debug("_gnerate_stroage_broker_lookup.key: {}".format(key))
+        storage_broker_lookup[key] = StorageBroker
+    logger.debug("_generate_storage_broker_lookup.return: {}".format(
+        storage_broker_lookup
+    ))
     return storage_broker_lookup
 
 
 def _get_storage_broker(uri, config_path):
     """Helper function to enable use lookup of appropriate storage brokers."""
+    logger.debug("In _get_storage_broker...")
     uri = dtoolcore.utils.sanitise_uri(uri)
     storage_broker_lookup = _generate_storage_broker_lookup()
+    logger.debug("_get_storage_broker.calling.utils.generous_parse_uri")
     parsed_uri = dtoolcore.utils.generous_parse_uri(uri)
     StorageBroker = storage_broker_lookup[parsed_uri.scheme]
-    return StorageBroker(uri, config_path)
+    storage_broker = StorageBroker(uri, config_path)
+    logger.debug("_get_storage_broker.return: {}".format(storage_broker))
+    return storage_broker
 
 
 def _admin_metadata_from_uri(uri, config_path):
     """Helper function for getting admin metadata."""
+    logger.debug("In _admin_metadata_from_uri...")
+    logger.debug("_admin_metadata_from_uri.input_uri: {}".format(uri))
     uri = dtoolcore.utils.sanitise_uri(uri)
     storage_broker = _get_storage_broker(uri, config_path)
     admin_metadata = storage_broker.get_admin_metadata()
+    logger.debug("_admin_metadata_from_uri.return: {}".format(admin_metadata))
     return admin_metadata
 
 
 def _is_dataset(uri, config_path):
     """Helper function for determining if a URI is a dataset."""
+    logger.debug("In _is_dataset...")
+    logger.debug("_is_dataset.input_uri: {}".format(uri))
     uri = dtoolcore.utils.sanitise_uri(uri)
     storage_broker = _get_storage_broker(uri, config_path)
     return storage_broker.has_admin_metadata()
@@ -68,6 +87,7 @@ def _is_dataset(uri, config_path):
 
 def generate_admin_metadata(name, creator_username=None):
     """Return admin metadata as a dictionary."""
+    logger.debug("In generate_admin_metadata...")
 
     if not dtoolcore.utils.name_is_valid(name):
         raise(DtoolCoreInvalidNameError())
@@ -95,13 +115,17 @@ def _generate_uri(admin_metadata, base_uri):
     :param base_uri: base URI from which to derive dataset URI
     :returns: dataset URI
     """
+    logger.debug("In _generate_uri...")
+    logger.debug("_generate_uri.input_base_uri: {}".format(base_uri))
     name = admin_metadata["name"]
     uuid = admin_metadata["uuid"]
     # storage_broker_lookup = _generate_storage_broker_lookup()
     # parse_result = urlparse(base_uri)
     # storage = parse_result.scheme
     StorageBroker = _get_storage_broker(base_uri, config_path=None)
-    return StorageBroker.generate_uri(name, uuid, base_uri)
+    uri = StorageBroker.generate_uri(name, uuid, base_uri)
+    logger.debug("_generate_uri.return: {}".format(uri))
+    return uri
 
 
 def generate_proto_dataset(admin_metadata, base_uri, config_path=None):
@@ -111,10 +135,65 @@ def generate_proto_dataset(admin_metadata, base_uri, config_path=None):
     :param base_uri: base URI for proto dataset
     :param config_path: path to dtool configuration file
     """
+    logger.debug("In generate_proto_dataset...")
     uri = _generate_uri(admin_metadata, base_uri)
     return ProtoDataSet(uri, admin_metadata, config_path)
 
 
+def create_proto_dataset(
+    name,
+    base_uri,
+    readme_content="",
+    creator_username=None
+):
+    """Return :class:`dtoolcore.ProtoDataSet` instance.
+
+    :param name: dataset name
+    :param base_uri: base URI for proto dataset
+    :param readme_content: content of README as a string
+    :param creator_username: creator username
+    """
+    logger.debug("In create_proto_dataset...")
+    admin_metadata = generate_admin_metadata(name, creator_username)
+    proto_dataset = generate_proto_dataset(admin_metadata, base_uri)
+    proto_dataset.create()
+    proto_dataset.put_readme(readme_content)
+    return proto_dataset
+
+
+def create_derived_proto_dataset(
+    name,
+    base_uri,
+    source_dataset,
+    readme_content="",
+    creator_username=None
+):
+    """Return :class:`dtoolcore.ProtoDataSet` instance.
+
+    It adds the "source_name", "source_uri", and "source_uuid" as
+    annotations.
+
+    :param name: dataset name
+    :param base_uri: base URI for proto dataset
+    :param source_dataset: source dataset
+    :param readme_content: content of README as a string
+    :param creator_username: creator username
+    """
+    logger.debug("In create_derived_proto_dataset...")
+    admin_metadata = generate_admin_metadata(name, creator_username)
+    proto_dataset = generate_proto_dataset(admin_metadata, base_uri)
+    proto_dataset.create()
+
+    # Add derived info as annotations.
+    proto_dataset.put_annotation("source_dataset_name", source_dataset.name)
+    proto_dataset.put_annotation("source_dataset_uri", source_dataset.uri)
+    proto_dataset.put_annotation("source_dataset_uuid", source_dataset.uuid)
+
+    proto_dataset.put_readme(readme_content)
+
+    return proto_dataset
+
+
 def _copy_create_proto_dataset(
     src_dataset,
     dest_base_uri,
@@ -259,11 +338,16 @@ class DtoolCoreInvalidNameError(ValueError):
     pass
 
 
+class DtoolCoreBrokenStagingPromise(IOError):
+    pass
+
+
 class _BaseDataSet(object):
     """Base class for datasets."""
 
     def __init__(self, uri, admin_metadata, config_path=None):
 
+        logger.debug("Initialising _BaseDataSet...")
         uri = dtoolcore.utils.sanitise_uri(uri)
 
         self._admin_metadata = admin_metadata
@@ -288,6 +372,7 @@ class _BaseDataSet(object):
     @property
     def base_uri(self):
         """Return the base URI of the dataset."""
+        logger.debug("In base_uri...")
         return self._storage_broker.generate_base_uri(self.uri)
 
     @property
@@ -372,13 +457,18 @@ class _BaseDataSet(object):
             "DTOOL_NUM_PROCESSES",
             default=1
         ))
-        logging.info(
-            "Using {} process(es) to generate manifest".format(
-                num_processes
-            )
-        )
 
-        if num_processes > 1:
+        # Using multiprocessing only makes sense for some storage brokers.
+        _mp_support = self._storage_broker.key == "file"  \
+            or self._storage_broker.key == "symlink"
+
+        if num_processes > 1 and _mp_support:
+
+            logger.info(
+                "Using {} process(es) to generate manifest".format(
+                    num_processes
+                )
+            )
 
             # Create pool of processes.
             pool = mp.Pool(num_processes)
@@ -446,6 +536,33 @@ class _BaseDataSet(object):
         logger.debug("List annotation names {}".format(self))
         return sorted(self._storage_broker.list_annotation_names())
 
+    def put_tag(self, tag):
+        """Annotate the dataset with a tag.
+
+        :param tag: tag
+        :raises: DtoolCoreInvalidNameError if the tag is invalid
+        :raises: DtoolCoreValueError if the tag is not a string
+        """
+        if not isinstance(tag, str):
+            raise(DtoolCoreValueError())
+
+        if not dtoolcore.utils.name_is_valid(tag):
+            raise(DtoolCoreInvalidNameError())
+
+        self._storage_broker.put_tag(tag)
+
+    def delete_tag(self, tag):
+        """Delete a tag from a dataset.
+
+        :param tag: tag
+        :raises: DtoolCoreKeyError if the tag does not exist
+        """
+        self._storage_broker.delete_tag(tag)
+
+    def list_tags(self):
+        """Return the dataset's tags as a list."""
+        return self._storage_broker.list_tags()
+
 
 class DataSet(_BaseDataSet):
     """
@@ -464,7 +581,7 @@ class DataSet(_BaseDataSet):
         """
         Return an existing :class:`dtoolcore.DataSet` from a URI.
 
-        :params uri: unique resource identifier where the existing
+        :param uri: unique resource identifier where the existing
                      :class:`dtoolcore.DataSet` is stored
         :returns: :class:`dtoolcore.DataSet`
         """
@@ -554,7 +671,7 @@ class ProtoDataSet(_BaseDataSet):
         """
         Return an existing :class:`dtoolcore.ProtoDataSet` from a URI.
 
-        :params uri: unique resource identifier where the existing
+        :param uri: unique resource identifier where the existing
                      :class:`dtoolcore.ProtoDataSet` is stored
         :returns: :class:`dtoolcore.ProtoDataSet`
         """
@@ -588,7 +705,7 @@ class ProtoDataSet(_BaseDataSet):
 
         :param fpath: path to the item on disk
         :param relpath: relative path name given to the item in the dataset as
-                        a handle
+                        a handle, i.e. a Unix-like relpath
         :returns: the handle given to the item
         """
         logger.debug("Put item with handle {} {}".format(relpath, self))
@@ -651,3 +768,201 @@ class ProtoDataSet(_BaseDataSet):
 
         # Clean up using the storage broker's post freeze hook.
         self._storage_broker.post_freeze_hook()
+
+
+class DataSetCreator(object):
+    """Context manager for creating a dataset.
+
+    Inside the context manager one works on a proto dataset.  When exiting the
+    context manager the proto dataset is automatically frozen into a dataset,
+    unless an exception has been raised in the context manager.
+    """
+
+    def __init__(
+        self,
+        name,
+        base_uri,
+        readme_content="",
+        creator_username=None
+    ):
+        base_uri = dtoolcore.utils.sanitise_uri(base_uri)
+        self.proto_dataset = create_proto_dataset(
+            name=name,
+            base_uri=base_uri,
+            readme_content=readme_content,
+            creator_username=creator_username
+        )
+        self._tmpdir = None
+
+    def __enter__(self):
+
+        self._to_stage = []
+
+        # Create the staging directory.
+        tmpdir_prefix = dtoolcore.utils.get_config_value(
+            "DTOOL_STAGING_PREFIX",
+            None
+        )
+        if tmpdir_prefix is None:
+            self._tmpdir = tempfile.mkdtemp()
+        else:
+            self._tmpdir = tempfile.mkdtemp(prefix=tmpdir_prefix)
+
+        return self
+
+    def __exit__(self, exception_type, exception_value, traceback):
+
+        # Add any staged files to the dataset.
+        for abspath, relpath in self._to_stage:
+            if not os.path.isfile(abspath):
+                raise(DtoolCoreBrokenStagingPromise(
+                    "No such file: {}".format(abspath)
+                ))
+            self.proto_dataset.put_item(abspath, relpath)
+
+        # If everything has been successful freeze the dataset.
+        if exception_type is None:
+            self.proto_dataset.freeze()
+
+        # Remove the staging directory.
+        shutil.rmtree(self._tmpdir)
+
+    @property
+    def name(self):
+        """Return the dataset name."""
+        return self.proto_dataset.name
+
+    @property
+    def uri(self):
+        """Return the dataset URI."""
+        return self.proto_dataset.uri
+
+    @property
+    def staging_directory(self):
+        """Return the staging directory.
+
+        An ephemeral directory that only exists within the
+        DataSetCreator context manger. It can be used as
+        a location to write output files that need to be
+        added to the dataset.
+
+        The easiest way to add a file here is to use the
+        :func:`dtoolcore.DataSetCreator.get_staging_fpath`
+        method to get a path to write content to.
+
+        If you write files directly to the staging directory
+        you will need to register them using the
+        :func:`dtoolcore.DataSetCreator.register_output_file`
+        method.
+        """
+        return self._tmpdir
+
+    def prepare_staging_abspath_promise(self, handle):
+        """Return abspath and handle to stage a file.
+
+        For getting access to an abspath that can be used to write output to.
+        It is the responsibility of this method to create any missing
+        subdirectories.  It is the responsibility of the user to create the
+        file associated with the abspath.
+
+        Use the abspath to create a file in the staging directory.
+        The file will be added to the dataset when exiting the
+        context handler.
+
+        The handle can be used to generate an identifier for the item in the
+        dataset using the :func:`dtoolcore.utils.generate_identifier` function.
+
+        :param handle: Unix like relpath
+        :returns: absolute path to the file in staging area
+                  that the user promises to create
+        """
+        _, ext = os.path.splitext(handle)
+        fname = dtoolcore.utils.generate_identifier(handle) + ext
+        staging_abspath = os.path.join(self.staging_directory, fname)
+
+        self._to_stage.append((staging_abspath, handle))
+
+        return staging_abspath
+
+    def put_item(self, fpath, relpath):
+        """
+        Put an item into the dataset.
+
+        :param fpath: path to the item on disk
+        :param relpath: relative path name given to the item in the dataset as
+                        a handle
+        :returns: the handle given to the item
+        """
+        return self.proto_dataset.put_item(fpath, relpath)
+
+    def add_item_metadata(self, handle, key, value):
+        """
+        Add metadata to a specific item in the :class:`dtoolcore.ProtoDataSet`.
+
+        :param handle: handle representing the relative path of the item in the
+                       :class:`dtoolcore.ProtoDataSet`
+        :param key: metadata key
+        :param value: metadata value
+        """
+        self.proto_dataset.add_item_metadata(handle, key, value)
+
+    def put_readme(self, content):
+        """
+        Update the README of the dataset and backup the previous README.
+
+        The client is responsible for ensuring that the content is valid YAML.
+
+        :param content: string to put into the README
+        """
+        self.proto_dataset.put_readme(content)
+
+    def put_annotation(self, annotation_name, annotation):
+        """Store annotation so that it is accessible by the given name.
+
+        :param annotation_name: name of the annotation
+        :param annotation: JSON serialisable value or data structure
+        :raises: DtoolCoreInvalidNameError if the annotation name
+                 is invalid
+        """
+        self.proto_dataset.put_annotation(annotation_name, annotation)
+
+    def put_tag(self, tag):
+        """Annotate the dataset with a tag.
+
+        :param tag: tag
+        :raises: DtoolCoreInvalidNameError if the tag is invalid
+        :raises: ValueError if the tag is not a string
+        """
+        self.proto_dataset.put_tag(tag)
+
+
+class DerivedDataSetCreator(DataSetCreator):
+    """Context manager for creating a derived dataset.
+
+    A derived dataset automatically has information about the source dataset
+    (name, URI and UUID) automatically added to the readme and to annotations.
+    It adds the "source_name", "source_uri", and "source_uuid" as annotations
+    and to the descriptive metadata in the readme.
+
+    Inside the context manager one works on a proto dataset.  When exiting the
+    context manager the proto dataset is automatically frozen into a dataset,
+    unless an exception has been raised in the context manager.
+    """
+
+    def __init__(
+        self,
+        name,
+        base_uri,
+        source_dataset,
+        readme_content="",
+        creator_username=None
+    ):
+        base_uri = dtoolcore.utils.sanitise_uri(base_uri)
+        self.proto_dataset = create_derived_proto_dataset(
+            name=name,
+            base_uri=base_uri,
+            source_dataset=source_dataset,
+            readme_content=readme_content,
+            creator_username=creator_username
+        )
+        self._tmpdir = None
